#include "difcalc.h"
#include "basefun.h"
#include "ordering.h"
#include <math.h>
#include <stdlib.h>

/**
   functions evaluates derivates of the function with variable xi
   with respect to x

   @param dx - array containing derivatives with respect to x
   @param jac - Jacobian
   @param x - array containing node coordinates
   @param xi - natural coordinate
   
   1.4.2002
*/
void derivatives_1d (vector &dx,double &jac,vector &x,double xi)
{
  long i;
  double d1,*nx;
  
  nx = new double [x.n];

  switch (x.n){
  case 2:{
    dx_bf_lin_1d (nx);
    break;
  }
  case 3:{
    dx_bf_quad_1d (nx,xi);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on element in function");
    fprintf (stderr,"\n derivatives_1d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  d1=0.0;
  for (i=0;i<x.n;i++){
    d1+=nx[i]*x[i];
  }
  
  jac = d1;
  
  for (i=0;i<dx.n;i++){
    dx[i]=nx[i]/jac;
  }
  
  delete [] nx;
}

/**
   functions evaluates jacobian

   @param jac - Jacobian
   @param x - array containing node coordinates
   @param xi - natural coordinate
   
   1.4.2002
*/
void jac_1d (double &jac,vector &x,double xi)
{
  long i;
  double d1,*nx;
  
  nx = new double [x.n];

  switch (x.n){
  case 2:{
    dx_bf_lin_1d (nx);
    break;
  }
  case 3:{
    dx_bf_quad_1d (nx,xi);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on element in function");
    fprintf (stderr,"\n jac_1d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  d1=0.0;
  for (i=0;i<x.n;i++){
    d1+=nx[i]*x[i];
  }
  
  jac = d1;
  
  delete [] nx;
}

/**
   functions evaluates derivates of the function with variables xi and eta
   with respect to x and y

   @param dx,dy - arrays containing derivatives with respect to x and y
   @param jac - Jacobian
   @param x,y - array containing node coordinates
   @param xi,eta - natural coordinates
   
   24.6.2001
*/
void derivatives_2d (vector &dx,vector &dy,double &jac,
		     vector &x,vector &y,double xi,double eta)
{
  long i;
  double d1,d2,d3,d4,ddx,ddy,*nx,*ny;
  
  nx = new double [x.n];
  ny = new double [x.n];

  switch (x.n){
  case 3:{
    dx_bf_lin_3_2d (nx);
    dy_bf_lin_3_2d (ny);
    break;
  }
  case 4:{
    dx_bf_lin_4_2d (nx,eta);
    dy_bf_lin_4_2d (ny,xi);
    break;
  }
  case 6:{
    dx_bf_quad_3_2d (nx,xi,eta);
    dy_bf_quad_3_2d (ny,xi,eta);
    break;
  }
  case 8:{
    dx_bf_quad_4_2d (nx,xi,eta);
    dy_bf_quad_4_2d (ny,xi,eta);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on 2D element in the function");
    fprintf (stderr,"\n derivatives_2d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  d1=0.0;  d2=0.0;	d3=0.0;  d4=0.0;
  for (i=0;i<x.n;i++){
    d1+=nx[i]*x[i];
    d2+=ny[i]*x[i];
    d3+=nx[i]*y[i];
    d4+=ny[i]*y[i];
  }
  
  jac = d1*d4 - d2*d3;
  
  for (i=0;i<dx.n;i++){
    ddx=dx[i];  ddy=dy[i];
    dx[i]=(ddx*d4-ddy*d3)/jac;
    dy[i]=(ddy*d1-ddx*d2)/jac;
  }
  delete [] nx;  delete [] ny;
}



/**
   functions evaluates Jacobian for 2D problems
   
   @param jac - Jacobian
   @param x,y - array containing node coordinates
   @param xi,eta - natural coordinates
   
   9.7.2001
*/
void jac_2d (double &jac,vector &x,vector &y,double xi,double eta)
{
  long i;
  double d1,d2,d3,d4,*nx,*ny;
  
  nx = new double [x.n];
  ny = new double [x.n];

  switch (x.n){
  case 3:{
    dx_bf_lin_3_2d (nx);
    dy_bf_lin_3_2d (ny);
    break;
  }
  case 4:{
    dx_bf_lin_4_2d (nx,eta);
    dy_bf_lin_4_2d (ny,xi);
    break;
  }
  case 6:{
    dx_bf_quad_3_2d (nx,xi,eta);
    dy_bf_quad_3_2d (ny,xi,eta);
    break;
  }
  case 8:{
    dx_bf_quad_4_2d (nx,xi,eta);
    dy_bf_quad_4_2d (ny,xi,eta);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on 2D element in the function");
    fprintf (stderr,"\n jac_2d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  d1=0.0;  d2=0.0;	d3=0.0;  d4=0.0;
  for (i=0;i<x.n;i++){
    d1+=nx[i]*x[i];
    d2+=ny[i]*x[i];
    d3+=nx[i]*y[i];
    d4+=ny[i]*y[i];
  }
  
  jac = d1*d4 - d2*d3;
  
  delete [] nx;  delete [] ny;
}



/**
   function evaluates derivatives of the function with variables xi, eta and zeta
   with respect to x, y and z
   
   @param dx,dy,dz - arrays containing derivatives with respect to x, y and z
   @param jac - Jacobian
   @param x,y,z - arrays containing node coordinates
   @param xi,eta,zeta - natural coordinates

   24.6.2001
*/
void derivatives_3d (vector &dx,vector &dy,vector &dz,double &jac,
		     vector &x,vector &y,vector &z,
		     double xi,double eta,double zeta)
{
  long i;
  double d1,d2,d3,d4,d5,d6,d7,d8,d9,ddx,ddy,ddz,*nx,*ny,*nz;
  
  nx = new double [x.n];
  ny = new double [x.n];
  nz = new double [x.n];

  switch (x.n){
  case 4:{
    dx_bf_lin_tet (nx);
    dy_bf_lin_tet (ny);
    dz_bf_lin_tet (nz);
    break;
  }
  case 6:{
    dx_bf_lin_wed_3d (nx,zeta);
    dy_bf_lin_wed_3d (ny,zeta);
    dz_bf_lin_wed_3d (nz,xi,eta);
    break;
  }
  case 8:{
    dx_bf_lin_hex_3d (nx,eta,zeta);
    dy_bf_lin_hex_3d (ny,xi,zeta);
    dz_bf_lin_hex_3d (nz,xi,eta);
    break;
  }
  case 10:{
    dx_bf_quad_tet(nx,xi,eta,zeta);
    dy_bf_quad_tet(ny,xi,eta,zeta);
    dz_bf_quad_tet(nz,xi,eta,zeta);
    break;
  }
  case 15:{
    dx_bf_quad_wed_3d (nx,xi,eta,zeta);
    dy_bf_quad_wed_3d (ny,xi,eta,zeta);
    dz_bf_quad_wed_3d (nz,xi,eta,zeta);
    break;
  }
  case 20:{
    dx_bf_quad_hex_3d (nx,xi,eta,zeta);
    dy_bf_quad_hex_3d (ny,xi,eta,zeta);
    dz_bf_quad_hex_3d (nz,xi,eta,zeta);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on 3D element in the function");
    fprintf (stderr,"\n derivatives_3d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  d1=0.0;  d2=0.0;  d3=0.0;  d4=0.0;
  d5=0.0;  d6=0.0;  d7=0.0;  d8=0.0;  d9=0.0;
  for (i=0;i<x.n;i++){
    d1+=nx[i]*x[i];
    d2+=ny[i]*x[i];
    d3+=nz[i]*x[i];
    d4+=nx[i]*y[i];
    d5+=ny[i]*y[i];
    d6+=nz[i]*y[i];
    d7+=nx[i]*z[i];
    d8+=ny[i]*z[i];
    d9+=nz[i]*z[i];
  }
  
  jac = d1*d5*d9 + d4*d8*d3 + d7*d2*d6 - d7*d5*d3 - d8*d6*d1 - d9*d4*d2;
  
  for (i=0;i<dx.n;i++){
    ddx=dx[i];  ddy=dy[i];  ddz=dz[i];
    dx[i]=(ddx*(d5*d9-d8*d6)+ddy*(d7*d6-d9*d4)+ddz*(d4*d8-d7*d5))/jac;
    dy[i]=(ddx*(d8*d3-d9*d2)+ddy*(d1*d9-d7*d3)+ddz*(d7*d2-d8*d1))/jac;
    dz[i]=(ddx*(d2*d6-d5*d3)+ddy*(d4*d3-d6*d1)+ddz*(d1*d5-d4*d2))/jac;
  }
  
  delete [] nx;  delete [] ny;  delete [] nz;
}



/**
   function evaluates Jacobian for 3D problems
   
   @param jac - Jacobian
   @param x,y,z - arrays containing node coordinates
   @param xi,eta,zeta - natural coordinates

   24.6.2001
*/
void jac_3d (double &jac,vector &x,vector &y,vector &z,
	     double xi,double eta,double zeta)
{
  long i;
  double d1,d2,d3,d4,d5,d6,d7,d8,d9,*nx,*ny,*nz;
  
  nx = new double [x.n];
  ny = new double [x.n];
  nz = new double [x.n];

  switch (x.n){
  case 4:{
    dx_bf_lin_tet (nx);
    dy_bf_lin_tet (ny);
    dz_bf_lin_tet (nz);
    break;
  }
  case 6:{
    dx_bf_lin_wed_3d (nx,zeta);
    dy_bf_lin_wed_3d (ny,zeta);
    dz_bf_lin_wed_3d (nz,xi,eta);
    break;
  }
  case 8:{
    dx_bf_lin_hex_3d (nx,eta,zeta);
    dy_bf_lin_hex_3d (ny,xi,zeta);
    dz_bf_lin_hex_3d (nz,xi,eta);
    break;
  }
  case 10:{
    dx_bf_quad_tet (nx,xi,eta,zeta);
    dy_bf_quad_tet (ny,xi,eta,zeta);
    dz_bf_quad_tet (nz,xi,eta,zeta);
    break;
  }
  case 15:{
    dx_bf_quad_wed_3d (nx,xi,eta,zeta);
    dy_bf_quad_wed_3d (ny,xi,eta,zeta);
    dz_bf_quad_wed_3d (nz,xi,eta,zeta);
    break;
  }
  case 20:{
    dx_bf_quad_hex_3d (nx,xi,eta,zeta);
    dy_bf_quad_hex_3d (ny,xi,eta,zeta);
    dz_bf_quad_hex_3d (nz,xi,eta,zeta);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on 3D element in the function");
    fprintf (stderr,"\n jac_3d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  d1=0.0;  d2=0.0;  d3=0.0;  d4=0.0;
  d5=0.0;  d6=0.0;  d7=0.0;  d8=0.0;  d9=0.0;
  for (i=0;i<x.n;i++){
    d1+=nx[i]*x[i];
    d2+=ny[i]*x[i];
    d3+=nz[i]*x[i];
    d4+=nx[i]*y[i];
    d5+=ny[i]*y[i];
    d6+=nz[i]*y[i];
    d7+=nx[i]*z[i];
    d8+=ny[i]*z[i];
    d9+=nz[i]*z[i];
  }
  
  jac = d1*d5*d9 + d4*d8*d3 + d7*d2*d6 - d7*d5*d3 - d8*d6*d1 - d9*d4*d2;

  delete [] nx;  delete [] ny;  delete [] nz;
}



/**
   function computes Jacobian of mapping from 2D curve to <-1;1> segment
   
   @param jac - Jacobian
   @param x,y - node coordinates
   @param xi - natural coordinate

   1.4.2002
*/
void jac1d2d (double &jac,vector &x,vector &y,double xi)
{
  long i,n;
  double dx,dy,*nx;
  
  n=x.n;
  nx = new double [n];
  
  switch (n){
  case 2:{
    dx_bf_lin_1d (nx);
    break;
  }
  case 3:{
    dx_bf_quad_1d (nx,xi);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on element in function");
    fprintf (stderr,"\n jac1d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  dx=0.0;  dy=0.0;
  for (i=0;i<n;i++){
    dx+=nx[i]*x[i];
    dy+=nx[i]*y[i];
  }
  
  jac = sqrt (dx*dx+dy*dy);
  
  delete [] nx;
}

/**
   function computes Jacobian of mapping from 3D curve to <-1;1> segment
   
   @param jac - Jacobian
   @param x,y,z - node coordinates
   @param xi - natural coordinate
   
   1.4.2002
*/
void jac1d3d (double &jac,vector &x,vector &y,vector &z,double xi)
{
  long i,n;
  double dx,dy,dz,*nx;
  
  n=x.n;
  nx = new double [n];
  
  switch (n){
  case 2:{
    dx_bf_lin_1d (nx);
    break;
  }
  case 3:{
    dx_bf_quad_1d (nx,xi);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on element in function");
    fprintf (stderr,"\n jac1d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  dx=0.0;  dy=0.0;  dz=0.0;
  for (i=0;i<n;i++){
    dx+=nx[i]*x[i];
    dy+=nx[i]*y[i];
    dz+=nx[i]*z[i];
  }
  
  jac = sqrt (dx*dx+dy*dy+dz*dz);
  
  delete [] nx;
}

/**
   function computes Jacobian of mapping from surface in 3D to unit square or triangle
   
   @param jac - jacobian
   @param x,y,z - arrays containing node coordinates
   @param xi,eta - natural coordinates
   
   1.4.2002
*/
void jac2d3d (double &jac,vector &x,vector &y,vector &z,double xi,double eta)
{
  long i,n;
  double dxu,dxv,dyu,dyv,dzu,dzv,g11,g22,g12;
  double *nu,*nv;
  
  n=x.n;
  nu = new double [n];  nv = new double [n];

  switch (n){
  case 3:{
    dx_bf_lin_3_2d (nu);
    dy_bf_lin_3_2d (nv);
    break;
  }
  case 4:{
    dx_bf_lin_4_2d (nu,eta);
    dy_bf_lin_4_2d (nv,xi);
    break;
  }
  case 6:{
    dx_bf_quad_3_2d (nu,xi,eta);
    dy_bf_quad_3_2d (nv,xi,eta);
    break;
  }
  case 8:{
    dx_bf_quad_4_2d (nu,xi,eta);
    dy_bf_quad_4_2d (nv,xi,eta);
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong number of nodes on 2D element in the function");
    fprintf (stderr,"\n jac2d3d (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  dxu=0.0;  dyu=0.0;  dzu=0.0;  dxv=0.0;  dyv=0.0;  dzv=0.0;
  for (i=0;i<n;i++){
    dxu+=nu[i]*x[i];
    dyu+=nu[i]*y[i];
    dzu+=nu[i]*z[i];
    dxv+=nv[i]*x[i];
    dyv+=nv[i]*y[i];
    dzv+=nv[i]*z[i];
  }
  
  g11=dxu*dxu+dyu*dyu+dzu*dzu;
  g22=dxv*dxv+dyv*dyv+dzv*dzv;
  g12=dxu*dxv+dyu*dyv+dzu*dzv;
  
  jac = sqrt (g11*g22-g12*g12);
  
  delete [] nu;  delete [] nv;
}

/**
   function computes jacobian of mapping from 2D curve to line <-1;1>
   curve is edge on 2D finite element
   
   @param jac - jacobian
   @param x,y - arrays containing node coordinates
   @param xi - natural coordinate
   @param edid - edge id
   
   1.4.2002
*/
void jac1d_2d (double &jac,vector &x,vector &y,double xi,long edid)
{
  vector lx,ly;
  
  if (x.n==3){
    allocv (2,lx);  allocv (2,ly);
    if (edid==0){
      lx[0]=x[0];  lx[1]=x[1];
      ly[0]=y[0];  ly[1]=y[1];
    }
    if (edid==1){
      lx[0]=x[1];  lx[1]=x[2];
      ly[0]=y[1];  ly[1]=y[2];
    }
    if (edid==2){
      lx[0]=x[2];  lx[1]=x[0];
      ly[0]=y[2];  ly[1]=y[0];
    }
  }

  if (x.n==4){
    allocv (2,lx);  allocv (2,ly);
    if (edid==0){
      lx[0]=x[0];  lx[1]=x[1];
      ly[0]=y[0];  ly[1]=y[1];
    }
    if (edid==1){
      lx[0]=x[1];  lx[1]=x[2];
      ly[0]=y[1];  ly[1]=y[2];
    }
    if (edid==2){
      lx[0]=x[2];  lx[1]=x[3];
      ly[0]=y[2];  ly[1]=y[3];
    }
    if (edid==3){
      lx[0]=x[3];  lx[1]=x[0];
      ly[0]=y[3];  ly[1]=y[0];
    }
  }
  
  if (x.n==6){
    allocv (3,lx);  allocv (3,ly);
    if (edid==0){
      lx[0]=x[0];  lx[1]=x[1];  lx[2]=x[3];
      ly[0]=y[0];  ly[1]=y[1];  ly[2]=y[3];
    }
    if (edid==1){
      lx[0]=x[1];  lx[1]=x[2];  lx[2]=x[4];
      ly[0]=y[1];  ly[1]=y[2];  ly[2]=y[4];
    }
    if (edid==2){
      lx[0]=x[2];  lx[1]=x[0];  lx[2]=x[5];
      ly[0]=y[2];  ly[1]=y[0];  ly[2]=y[5];
    }
  }

  if (x.n==8){
    allocv (3,lx);  allocv (3,ly);
    if (edid==0){
      lx[0]=x[0];  lx[1]=x[1];  lx[2]=x[4];
      ly[0]=y[0];  ly[1]=y[1];  ly[2]=y[4];
    }
    if (edid==1){
      lx[0]=x[1];  lx[1]=x[2];  lx[2]=x[5];
      ly[0]=y[1];  ly[1]=y[2];  ly[2]=y[5];
    }
    if (edid==2){
      lx[0]=x[2];  lx[1]=x[3];  lx[2]=x[6];
      ly[0]=y[2];  ly[1]=y[3];  ly[2]=y[6];
    }
    if (edid==3){
      lx[0]=x[3];  lx[1]=x[0];  lx[2]=x[7];
      ly[0]=y[3];  ly[1]=y[0];  ly[2]=y[7];
    }
  }

  jac1d2d (jac,lx,ly,xi);
  
  destrv (lx);  destrv (ly);
}

/**
   function computes jacobian of mapping from surface to unit square or triangle
   surface is surface of 3D finite element
   
   @param jac - jacobian
   @param x,y,z - node coordinates
   @param xi,eta - natural coordinates
   @param sid - surface id
   
*/
void jac2d_3d (double &jac,vector &x,vector &y,vector &z,double xi,double eta,long sid)
{
  vector lx,ly,lz;
  ivector sn;
  long i;
  
  if (x.n==4){
    // linear tetraherdon
    allocv (3,lx);  allocv (3,ly);  allocv (3,lz);
    allocv (3,sn);
    if (sid==0){
      /*
      lx[0]=x[1];  lx[1]=x[2];  lx[2]=x[3];
      ly[0]=y[1];  ly[1]=y[2];  ly[2]=y[3];
      lz[0]=z[1];  lz[1]=z[2];  lz[2]=z[3];
      */
      lx[0]=x[2];  lx[1]=x[1];  lx[2]=x[3];
      ly[0]=y[2];  ly[1]=y[1];  ly[2]=y[3];
      lz[0]=z[2];  lz[1]=z[1];  lz[2]=z[3];
    }
    if (sid==1){
      lx[0]=x[0];  lx[1]=x[2];  lx[2]=x[3];
      ly[0]=y[0];  ly[1]=y[2];  ly[2]=y[3];
      lz[0]=z[0];  lz[1]=z[2];  lz[2]=z[3];
    }
    if (sid==2){
      lx[0]=x[1];  lx[1]=x[0];  lx[2]=x[3];
      ly[0]=y[1];  ly[1]=y[0];  ly[2]=y[3];
      lz[0]=z[1];  lz[1]=z[0];  lz[2]=z[3];
    }
    if (sid==3){
      lx[0]=x[0];  lx[1]=x[1];  lx[2]=x[2];
      ly[0]=y[0];  ly[1]=y[1];  ly[2]=y[2];
      lz[0]=z[0];  lz[1]=z[1];  lz[2]=z[2];
    }
  }

  if (x.n==10)
  {
    // quadratic tetrahedron
    allocv (6,lx);  allocv (6,ly);  allocv (6,lz);
    allocv (6,sn);
    quadtetrahedral_surfnod (sn.a,sid);
    for (i=0; i<6; i++)
    {
      lx[i] = x[sn[i]];
      ly[i] = y[sn[i]];
      lz[i] = z[sn[i]];
    }
  }


  if (x.n==8){
    // linear hexahedron
    allocv (4,lx);  allocv (4,ly);  allocv (4,lz);
    allocv (4,sn);
    if (sid==0){
      lx[0]=x[0];  lx[1]=x[3];  lx[2]=x[7];  lx[3]=x[4];
      ly[0]=y[0];  ly[1]=y[3];  ly[2]=y[7];  ly[3]=y[4];
      lz[0]=z[0];  lz[1]=z[3];  lz[2]=z[7];  lz[3]=z[4];
    }
    if (sid==1){
      lx[0]=x[1];  lx[1]=x[0];  lx[2]=x[4];  lx[3]=x[5];
      ly[0]=y[1];  ly[1]=y[0];  ly[2]=y[4];  ly[3]=y[5];
      lz[0]=z[1];  lz[1]=z[0];  lz[2]=z[4];  lz[3]=z[5];
    }
    if (sid==2){
      lx[0]=x[2];  lx[1]=x[1];  lx[2]=x[5];  lx[3]=x[6];
      ly[0]=y[2];  ly[1]=y[1];  ly[2]=y[5];  ly[3]=y[6];
      lz[0]=z[2];  lz[1]=z[1];  lz[2]=z[5];  lz[3]=z[6];
    }
    if (sid==3){
      lx[0]=x[3];  lx[1]=x[2];  lx[2]=x[6];  lx[3]=x[7];
      ly[0]=y[3];  ly[1]=y[2];  ly[2]=y[6];  ly[3]=y[7];
      lz[0]=z[3];  lz[1]=z[2];  lz[2]=z[6];  lz[3]=z[7];
    }
    if (sid==4){
      lx[0]=x[0];  lx[1]=x[1];  lx[2]=x[2];  lx[3]=x[3];
      ly[0]=y[0];  ly[1]=y[1];  ly[2]=y[2];  ly[3]=y[3];
      lz[0]=z[0];  lz[1]=z[1];  lz[2]=z[2];  lz[3]=z[3];
    }
    if (sid==5){
      lx[0]=x[4];  lx[1]=x[7];  lx[2]=x[6];  lx[3]=x[5];
      ly[0]=y[4];  ly[1]=y[7];  ly[2]=y[6];  ly[3]=y[5];
      lz[0]=z[4];  lz[1]=z[7];  lz[2]=z[6];  lz[3]=z[5];
    }
  }
  
  if (x.n==20){
    // quadratic hexahedron
    allocv (8,lx);  allocv (8,ly);  allocv (8,lz);
    allocv (8,sn);
    if (sid==0){
      lx[0]=x[0];  lx[1]=x[3];  lx[2]=x[7];  lx[3]=x[4];   lx[4]=x[11];  lx[5]=x[15];  lx[6]=x[19];  lx[7]=x[12];
      ly[0]=y[0];  ly[1]=y[3];  ly[2]=y[7];  ly[3]=y[4];   ly[4]=y[11];  ly[5]=y[15];  ly[6]=y[19];  ly[7]=y[12];
      lz[0]=z[0];  lz[1]=z[3];  lz[2]=z[7];  lz[3]=z[4];   lz[4]=z[11];  lz[5]=z[15];  lz[6]=z[19];  lz[7]=z[12];
    }
    if (sid==1){
      lx[0]=x[1];  lx[1]=x[0];  lx[2]=x[4];  lx[3]=x[5];   lx[4]=x[8];   lx[5]=x[12];  lx[6]=x[16];  lx[7]=x[13];
      ly[0]=y[1];  ly[1]=y[0];  ly[2]=y[4];  ly[3]=y[5];   ly[4]=y[8];   ly[5]=y[12];  ly[6]=y[16];  ly[7]=y[13];
      lz[0]=z[1];  lz[1]=z[0];  lz[2]=z[4];  lz[3]=z[5];   lz[4]=z[8];   lz[5]=z[12];  lz[6]=z[16];  lz[7]=z[13];
    }
    if (sid==2){
      lx[0]=x[2];  lx[1]=x[1];  lx[2]=x[5];  lx[3]=x[6];   lx[4]=x[9];   lx[5]=x[13];  lx[6]=x[17];  lx[7]=x[14];
      ly[0]=y[2];  ly[1]=y[1];  ly[2]=y[5];  ly[3]=y[6];   ly[4]=y[9];   ly[5]=y[13];  ly[6]=y[17];  ly[7]=y[14];
      lz[0]=z[2];  lz[1]=z[1];  lz[2]=z[5];  lz[3]=z[6];   lz[4]=z[9];   lz[5]=z[13];  lz[6]=z[17];  lz[7]=z[14];
    }
    if (sid==3){
      lx[0]=x[3];  lx[1]=x[2];  lx[2]=x[6];  lx[3]=x[7];   lx[4]=x[10];  lx[5]=x[14];  lx[6]=x[18];  lx[7]=x[15];
      ly[0]=y[3];  ly[1]=y[2];  ly[2]=y[6];  ly[3]=y[7];   ly[4]=y[10];  ly[5]=y[14];  ly[6]=y[18];  ly[7]=y[15];
      lz[0]=z[3];  lz[1]=z[2];  lz[2]=z[6];  lz[3]=z[7];   lz[4]=z[10];  lz[5]=z[14];  lz[6]=z[18];  lz[7]=z[15];
    }
    if (sid==4){
      lx[0]=x[0];  lx[1]=x[1];  lx[2]=x[2];  lx[3]=x[3];   lx[4]=x[8];   lx[5]=x[9];   lx[6]=x[10];  lx[7]=x[11];
      ly[0]=y[0];  ly[1]=y[1];  ly[2]=y[2];  ly[3]=y[3];   ly[4]=y[8];   ly[5]=y[9];   ly[6]=y[10];  ly[7]=y[11];
      lz[0]=z[0];  lz[1]=z[1];  lz[2]=z[2];  lz[3]=z[3];   lz[4]=z[8];   lz[5]=z[9];   lz[6]=z[10];  lz[7]=z[11];
    }
    if (sid==5){
      lx[0]=x[4];  lx[1]=x[7];  lx[2]=x[6];  lx[3]=x[5];   lx[4]=x[19];  lx[5]=x[18];  lx[6]=x[17];  lx[7]=x[16];
      ly[0]=y[4];  ly[1]=y[7];  ly[2]=y[6];  ly[3]=y[5];   ly[4]=y[19];  ly[5]=y[18];  ly[6]=y[17];  ly[7]=y[16];
      lz[0]=z[4];  lz[1]=z[7];  lz[2]=z[6];  lz[3]=z[5];   lz[4]=z[19];  lz[5]=z[18];  lz[6]=z[17];  lz[7]=z[16];
    }
  }


  jac2d3d (jac,lx,ly,lz,xi,eta);
  
  destrv (lx);  destrv (ly);  destrv (lz);
  destrv (sn);
}

