/*
    Copyright (C) 2003  David Masin
    email: masin@natur.cuni.cz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation 
    59 Temple Place, Suite 330, Boston, MA, 02111-1307, USA

    The use of the program is a subject to acknowledgement
*/

#include <iostream>   
#include <fstream> 
#include <cstring>
#include <cmath>
#include <vector>
#include <cassert> 
#include "matrix.h"
#include "generalmod.h"
#include <cstdlib>

using namespace std;

const int MDIM=3;
const int MCHAR=100;
const double EULER = 2.718281828459045235;
const double TINY = 1.e-20;
const double PI=3.14159265358979323846264338327050288;
const double onethird=0.33333333333333333333333333333333333333333333333333333333333333333333;
const double small_diff=0.00001; //Ensures that stress state is not inside 
				 //y & h surface - program would calculate elastically

//Only one of CLOE models possible:
#define CLOE_USE 0
#define CLOE_CLAY_USE 0

/*******************FUNKCE PRO MATEMATICKE OPERACE********************************/

void array_add( double a[], double b[], double c[], long int n );
double array_inproduct( double a[], double b[], long int n ); 
void array_move( double from[], double to[], long int n );
void array_multiply( double a[], double b[], double c, long int n );
void array_set( double *ptr, double value, long int n );
double array_size( double a[], long int n ); 
void array_subtract( double a[], double b[], double c[], long int n );
void matrix_eigenvalues( double mat[], double eigenvalues[] );
double scalar_dabs(double k);
void matrix_invariants( double *mat, double *inv );
void matrix_inverse(double matr[], double inv[], int N);
double matrix_determinant( double a[], long int n );
void make_dev(double tnz[9], double dev[9]);
double giverad(double deg);
double givedeg(double rad);
double scalar_power( double a, double b );
double scalar_square( double a );
bool string_isdouble( string word );
double trace( double a[3][3]);
void matrix4_ab( double a[], double b[], double c[3][3][3][3] );
void matrix_ab4( double a[], double b[3][3][3][3], double c[] );
void matrix_a4b( double a[3][3][3][3],  double b[], double c[]);
void matrix_a_contr_b( double a[], double b[], double &c );
void matrix_ab( double *a, double *b, double *c, long int n, long int m, long int k);
void make_kron_delta(double kron_delta[MDIM][MDIM]);
int max(vector<int> vektor);
int min(vector<int> vektor);
/* in file tskh.cc */
void calc_IJlode(double geosigma[9], double &I, double &J, double &lode, bool calcderiv, 
	double dIdsig[3][3], double dJdsig[3][3], double dlodedsig[3][3]);

/*************************Conditions for test controll************************************/

enum Condition {
	epax,
	epsh,
	epv,
	p,
	q,
	sigax,
	lode,
	eta,
	ep12,
	scycle,
	suction,
	temper
};

enum Control_mode {
	no_spec,
	strain,
	strain_auto,
	strain_rad,
	strain_ax,
	stress,

	strain_0,
	strain_4,
	strain_8,
	stress_0,
	stress_4,
	stress_8
};

enum Test_type {
	not_spec,
	constp3D,
	cpcirc3D
};

enum Res_type {
	results3D,
	results2D,
	strainenergy,
	unsat
};

enum Hypo_mode {
	basic,
	DdepL,
	istrain,
	istrain_anis,
	istrain_unsat_anis,
	istrain_unsat_Sr_anis,
	istrain_pdep,
	barodesymode
};

enum Model_type {
	hypoplasticity,
	elastoplasticity,
	umatmodel,
	umatunsatmodel,
	generalmodel
};

enum Model_name {
	hypo_clay_anisot,
	hypo_clay_explicit,
	hypo_clay_unsat_anisot,
	other_model
};

/*************************PROMENNE*******************************************/

struct Variable {
	double  sig[MDIM*MDIM], new_sig[MDIM*MDIM], old_sig[MDIM*MDIM],	 	
	 	epe[MDIM*MDIM], new_epe[MDIM*MDIM],
		epp[MDIM*MDIM], new_epp[MDIM*MDIM],
		geo_sig[MDIM*MDIM], geo_epe[MDIM*MDIM], geo_epp[MDIM*MDIM],
		d_geo_epe[MDIM*MDIM], d_geo_epp[MDIM*MDIM], d_epp[MDIM*MDIM], d_epe[MDIM*MDIM];
	double ua, uw, new_ua, new_uw, temperature, new_temperature, Sr, new_Sr, time, new_time;
	double geoept[9], geoepts, geoeptv, geop, geoq, K, G, Gsimpleshear,loderounddeg, lodeeptrounddeg, W;
	double init_psidir, last_psidir, scycle_centre[9];
	int scycle_num, scycle_max, nparms, nstatev, ngstrain, ngstress, ngstrainmax, ngstressmax;
	double delta_sig_trial[MDIM*MDIM], d_ua_trial, d_uw_trial;
	double df_trial[MDIM*MDIM], dg[MDIM*MDIM], df_ds_trial;
	double kron_delta[MDIM][MDIM];
	double e_perm[MDIM][MDIM][MDIM];
	double matrix_Cep[MDIM][MDIM][MDIM][MDIM], Kp_unsat_SI, Kp_unsat_LC;
	double matrix_Cep_inv[MDIM][MDIM][MDIM][MDIM];
	double matrix_C[MDIM][MDIM][MDIM][MDIM], Ke_unsat;
	double matrix_C_inv[MDIM][MDIM][MDIM][MDIM];
	double hypo_L[MDIM][MDIM][MDIM][MDIM];
	double hypo_Le_unsat[MDIM][MDIM][MDIM][MDIM];
	double Ristr,lambdap_su_inv,lambdafact,lambda_act; 
	double hypo_L_unl[MDIM][MDIM][MDIM][MDIM], H_unsat[MDIM][MDIM], HT_unsat[MDIM][MDIM];
	double hypo_N[MDIM][MDIM], fd;
	double barodR[MDIM][MDIM], baroda, barodb;
	double matrix_I[MDIM][MDIM][MDIM][MDIM];
	double f_old, f, max_f, min_f, g, max_f2, min_f2;
	Condition condition;
	bool first_stage, plasti, strdir, store_results;
	Res_type restype;
	vector<double> elasti_data;
	vector<double> plasti_data;
	vector<double> triax_data;
	vector<double> hisv;
        vector<double> new_hisv;
	vector<double> extv;
	vector<double> old_extv;
	vector<double> old_hisv;
	vector<double> results;		
	vector<double> results_ys;
	vector<double> results_ps;
	vector<double> results_os;
	vector<double> results_3Ds;
	vector<double> results_re;
	vector<double> results_re_special;
	vector<double> results_hypo_sbs;	
	vector<double> results_hypo_sbs_special;
	vector<double> other_f;
	vector<double> dother_f_ds_trial;
	vector<bool> plasti_unsat;
	int saveevery;

	Variable() {
		inicialize();
		condition=epax;
		store_results=true;
		restype=results2D;
		saveevery=1;
	}
	void inicialize();
	void recount(bool noGK); 
};

/*************************ELASTICITA*******************************************/

struct Elasti_rule {                        					//Rodicovska trida pro vsechny elasticke modely
	vector<double> elasti_data;						//Pocet parametru (kvuli modelum s predem neznamym poctem parametru)	
	bool tris;
	Elasti_rule (bool tr=false) : tris(tr){}
	virtual void give_matrix(bool trial) {cout<<"Elasti nefunguje virtualne ";};						//Fce vytvori matici tuhosti pro dany model 
	virtual void make_matrix_C (double young, double poisson,		//Fce obecne sestavi matici tuhosti pri zadani parametru	
		 double C[MDIM][MDIM][MDIM][MDIM]); 
	virtual void make_matrix_C_tris_grahoul (double young, double poisson, double alpha, 
		 double C[MDIM][MDIM][MDIM][MDIM]); 
};

struct Young : Elasti_rule {							//Potomek Elasti_rule - Younguv modul konstantni
	Young () : Elasti_rule() {}; 
	void give_matrix(bool trial);
};

struct Young_power : Elasti_rule {						//Potomek Elasti_rule - Younguv modul zavisly na napeti
	Young_power () : Elasti_rule() {}; 
	void give_matrix(bool trial);
};

struct Young_strainstress : Elasti_rule {
	Young_strainstress () : Elasti_rule() {}; 
	void give_matrix(bool trial);
};

struct Young_tris : Elasti_rule {
	Young_tris () : Elasti_rule() {}; 
	void give_matrix(bool trial);
};
struct Smallstrain_elasti : Elasti_rule {				
	Smallstrain_elasti (bool tr=false) : Elasti_rule(tr) {}; 
	void give_matrix(bool trial);
};

struct Hypoelasti : Elasti_rule {
	Hypoelasti () : Elasti_rule() {}; 
	void give_matrix(bool trial);
};

struct Severntrent_elasti : Elasti_rule {
	Severntrent_elasti () : Elasti_rule() {}; 
	void give_matrix(bool trial);
};

struct Camclay_elasti : Elasti_rule {
	Camclay_elasti () : Elasti_rule() {}; 
	virtual void give_matrix(bool trial);
	virtual bool loglog() {return(false);};
	virtual bool presdepG() {return(false);};
	virtual bool constpoisson() {return(false);};
};

struct Barcelona_basic_elasti : Elasti_rule {
	Barcelona_basic_elasti () : Elasti_rule() {}; 
	virtual void give_matrix(bool trial);
};

struct Camclay_loglog_presdepG_elasti : Camclay_elasti {
	Camclay_loglog_presdepG_elasti () : Camclay_elasti() {}; 
	bool presdepG() {return(true);};
	bool loglog() {return(true);};
};

struct Camclay_poisson_elasti : Camclay_elasti {
	Camclay_poisson_elasti () : Camclay_elasti() {}; 
	bool presdepG() {return(false);};
	bool loglog() {return(false);};
	bool constpoisson() {return(true);};
};

struct Camclay_loglog_elasti : Camclay_elasti {
	Camclay_loglog_elasti () : Camclay_elasti() {}; 
	bool loglog() {return(true);};
};

struct Gklinp_elasti : Elasti_rule {
	Gklinp_elasti () : Elasti_rule() {}; 
	virtual void give_matrix(bool trial);
};

struct Tskh_elasti : Elasti_rule {
	Tskh_elasti (bool tr=false) : Elasti_rule(tr) {}; 
	void give_matrix(bool trial);
};

/*****************************PLASTICITA***************************************/

struct Plasti_rule {         							//Rodicovska trida pro vsechny podminky plasticity
	Model_type model_type;
	Model_name model_name;
	bool unsat, Tdepse, addstrain;
	Hypo_mode hypo_mode;
	int add_for_istrainparam;
	double step;
	double eppnorm, signorm;
	double sairentry, gamma, lambdap, alpha_s, beta;
	Plasti_rule (Model_type hu=elastoplasticity, Hypo_mode istr=basic, int add=0, Model_name mn=other_model) : model_type(hu), hypo_mode(istr), add_for_istrainparam(add), model_name(mn) {eppnorm=10; signorm=100; unsat=false; addstrain=false;};
	virtual void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv) {};
	virtual void give_matrix(bool only_f) {cout<<"Plasti nefunguje virtualne ";};						
	virtual void give_matrix_general(bool only_f);						
	virtual void make_matrix_Cep (double df[], double dg[], double H);
	virtual void inic_hisv() {};
	virtual void give_LN() {cout<<"virtual hypoplasti give_LN, something wrong"; exit(0);};
	virtual void give_barodesy_Rab(double Dstretch[3][3]) {cout<<"virtual give_barodesy_Rab, something wrong"; exit(0);};
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void num_hisv(int &num_hisv) {num_hisv=1;};
	virtual void make_sig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double tensor_unsat[MDIM*MDIM], double &scalar_unsat);
	virtual void make_dsig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double dsig[MDIM*MDIM], double dua, double duw, 
		double dtensor_unsat[MDIM*MDIM], double &dscalar_unsat);
	virtual void calc_depm_suction(double delta_sig[9], double d_ua, double d_uw,double delta_epadd[9],double &factor) {array_set(delta_epadd,0,9);factor=0;};
	virtual void calc_depm_thermal(double delta_temperature,double delta_ep_thermal[9],double &factor) {array_set(delta_ep_thermal,0,9);};
	virtual void calc_depm(double delta_sig[9], double d_ua, double d_uw,double delta_temperature,double delta_epadd[9],double &factor){array_set(delta_epadd,0,9);factor=0;};
	virtual bool unsat_stress_varWRC_correction(double dep_Meef[9], double delta_temper, double &correction) {return(false);};
};

struct Mohr_Coulomb : Plasti_rule {						//Potomek Plasti_rule - Standardni Mohr-Coulombuv model
	Mohr_Coulomb () : Plasti_rule() {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
};

struct Mohrcoul_softening : Plasti_rule {
	Mohrcoul_softening () :	 Plasti_rule() {}; 
	void inic_hisv(); 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
};

struct Druckprag : Plasti_rule {					
	Druckprag () : Plasti_rule() {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
};

struct Lagpotts_camclay_plasti : Plasti_rule {					
	Lagpotts_camclay_plasti () : Plasti_rule() {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
	void inic_hisv();
};

struct Lagpotts_cem : Plasti_rule {					
	Lagpotts_cem () : Plasti_rule() {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
	void inic_hisv();
};

struct Camclay_plasti : Plasti_rule {					
	Camclay_plasti () : Plasti_rule() {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
	void inic_hisv();
};

struct Barcelona_basic_plasti : Plasti_rule {					
	double Gparam, Nparam, lambda0, beta, rparam, kappa, pc, lambdas, kappas, Mparam, kparam, pat;
	Barcelona_basic_plasti () : Plasti_rule() {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
	void inic_hisv();
	virtual void make_sig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double tensor_unsat[MDIM*MDIM], double &scalar_unsat);
};

struct Camclay_soft_plasti : Plasti_rule {					
	double G, m, kappa, lambda, N, sf, kparam, Aparam;
	Camclay_soft_plasti () : Plasti_rule() {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
	void inic_hisv();
};

enum Tskh_mode {
	tskh,
	tskh_skew,
	aitskh,
	aitskh_allhisv
};

struct Tskh_plasti : Plasti_rule {	
	Tskh_mode mode;
	double m, kappa, lambda, T, S, psi, N, m_flow, r_m, r_mfl;
	double beta, first_part, second_part; 				
	double dbetadepp[9];
	Tskh_plasti (Tskh_mode m) : Plasti_rule(), mode(m) {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void calcandderiv_beta();
	void give_matrix(bool only_f);
	void inic_hisv();
};

struct Tskh_general_plasti : Plasti_rule {	
	Tskh_mode mode;
	double m, kappa, lambda, T, S, psi, N, m_flow, r_m, r_mfl, m_cmp, df[9], m_flow_K0,
		rolodeb, rolodefl, drolodenbdlodenb, drolodefldlodenb;
	double sig_b[9], sig_a[9], I_0, sig_0[9];
	Tskh_general_plasti (Tskh_mode m) : Plasti_rule(), mode(m){}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
	void inic_hisv();
	void calc_rolode(double lode, double beta, double &rolode, double &drolodedlode, 
		bool alsoderivative);
};

struct Severntrent_plasti : Plasti_rule {	
	double phic, lambda, nu_lambda, Rparam, Bparam, kparam, Aparam, kdparam;
	Severntrent_plasti () : Plasti_rule() {}; 
	void give_f(double sig[MDIM*MDIM], double epp[9], double d_epp[9],
		 double &f, double &g, bool recount_hisv);
	void give_matrix(bool only_f);
	void inic_hisv();
};

/*_______________________________Hypoplasticity____________________________________*/

struct Hypoplasti_clay : Plasti_rule {			
	double ficr, rparam, N, lambda, kappa;
	Hypoplasti_clay(Hypo_mode istr, int add=-3) : Plasti_rule(hypoplasticity, istr, add){}; 
	virtual void inic_hisv(); 
	virtual void give_LN();
	virtual void calc_N(double &Nuse, double &rparam_use, double &kappa_use) {};
	virtual void calc_fs(double &fs, double trT, double a, double fdi) {};
	void rotate_L(double T[3][3], double hypo_m[3][3]);
};

struct Hypoplasti_clay_explicit : Plasti_rule {			
	double phic, nuparam, Nparam, lambda, kappa;
	Hypoplasti_clay_explicit(Hypo_mode istr, int add=-3, Model_name model_name=hypo_clay_explicit) : Plasti_rule(hypoplasticity, istr, add, model_name){}; 
	virtual void inic_hisv(); 
	virtual void give_LN();
};

struct Hypoplasti_clay_anisot : Plasti_rule {			
	double phic, nuparam, Nparam, lambda, kappa, alphanu, alphaE, alphaG;
	Hypoplasti_clay_anisot(Hypo_mode istr, int add=-2, Model_name model_name=hypo_clay_anisot) : Plasti_rule(hypoplasticity, istr, add, model_name){}; 
	virtual void inic_hisv(); 
	virtual void give_LN();
};

struct Hypoplasti_camclay : Plasti_rule {			
	double Mparam, nuparam, Nparam, lambda, kappa;
	Hypoplasti_camclay(Hypo_mode istr, int add=-3) : Plasti_rule(hypoplasticity, istr, add){}; 
	virtual void inic_hisv(); 
	virtual void give_LN();
};

struct Hypoplasti_clay_soft : Hypoplasti_clay {			
	double sf, kparam, Aparam;
	Hypoplasti_clay_soft(Hypo_mode istr) : Hypoplasti_clay(istr,0){}; 
	virtual void calc_N(double &Nuse, double &rparam_use, double &kappa_use);
	virtual void calc_fs(double &fs, double trT, double a, double fdi);
	virtual void inic_hisv(); 
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void num_hisv(int &num_hisv) {num_hisv=2;};
};

struct Hypoplasti_huang : Plasti_rule {			
	double ficr, ri, nparam, lambda, kappa;
	Hypoplasti_huang(Hypo_mode istr, int add=-3) : Plasti_rule(hypoplasticity, istr, add){}; 
	virtual void inic_hisv(); 
	virtual void give_LN();
};

struct Hypoplasti_lowangles : Plasti_rule {			
	bool hypo_presdepe;		
	double ficr, hs, nparam, ec0, ed0, ei0, alpha, beta, rparam, powxi;
	Hypoplasti_lowangles(Hypo_mode istr, bool pre) : Plasti_rule(hypoplasticity, istr, 2), hypo_presdepe(pre){}; 
	virtual void inic_hisv(); 
	virtual void give_LN();
};

struct Hypoplasti_vW : Plasti_rule {			
	bool hypo_presdepe;		
	double ficr, hs, nparam, ec0, ed0, ei0, alpha, beta;
	Hypoplasti_vW(Hypo_mode istr, bool pre, int add=0) : Plasti_rule(hypoplasticity, istr, add), hypo_presdepe(pre){}; 
	virtual void inic_hisv(); 
	virtual void give_LN();
	virtual void calc_fs(double &fs, double trT, double a) {};
	virtual void give_s(double &s_use) {};
};

struct Hypoplasti_vW_soft : Hypoplasti_vW {			
	double sf, kparam, Aparam;
	Hypoplasti_vW_soft(Hypo_mode istr, bool pre) : Hypoplasti_vW(istr, pre, 3){}; 
	virtual void give_s(double &s_use);
	virtual void calc_fs(double &fs, double trT, double a);
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void inic_hisv(); 
};

struct Hypoplasti_clay_unsat : Plasti_rule {			
	double ficr, rparam, N, lambda, kappa;
	double CSparam, mparam, nparam, lparam, sairentry0, lambdap0, lambdap_su,lambdapuse;
	double stresssuction,pdivpsbs;
	bool edepse, m0_OCRindep_collapse, expansive;
	Hypoplasti_clay_unsat(Hypo_mode istr, int add=2) : Plasti_rule(hypoplasticity, istr, add){
		edepse=false;
		Tdepse=false;
		m0_OCRindep_collapse=false;
		expansive=false;
	}; 
	virtual void calc_N_lambda_dpeds(double &Nuse, double &lambda_use, double &dpeds_divpe, double &dpedT_divpe);
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void inic_hisv(); 
	virtual void give_LN();
	virtual void make_sig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double tensor_unsat[MDIM*MDIM], double &scalar_unsat);
	virtual void make_dsig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double dsig[MDIM*MDIM], double dua, double duw, 
		double dtensor_unsat[MDIM*MDIM], double &dscalar_unsat);
	virtual void calc_sairentry(bool init) {sairentry=sairentry0;};
};

struct Hypoplasti_clay_unsat_anisot : Hypoplasti_clay_unsat {			
	double phic, nuparam, alphanu, alphaE, alphaG, se, lambdafact;
	Hypoplasti_clay_unsat_anisot(Hypo_mode istr, int add=3) : Hypoplasti_clay_unsat(istr, add){}; 
	virtual void inic_hisv();
	virtual void give_LN();
};

struct Hypoplasti_clay_unsat_Sr_anisot : Hypoplasti_clay_unsat_anisot {			
	double e0,aer;
	int scanning;
	Hypoplasti_clay_unsat_Sr_anisot(Hypo_mode istr, int add=6) : Hypoplasti_clay_unsat_anisot(istr, add){}; 
	virtual void calc_N_lambda_dpeds(double &Nuse, double &lambda_use, double &dpeds_divpe, double &dpedT_divpe);
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void inic_hisv(); 
	virtual void make_sig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double tensor_unsat[MDIM*MDIM], double &scalar_unsat);
	virtual void make_dsig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double dsig[MDIM*MDIM], double dua, double duw, 
		double dtensor_unsat[MDIM*MDIM], double &dscalar_unsat);	
	virtual void calc_sairentry(bool init);
	virtual void num_hisv(int &num_hisv) {num_hisv=3;};
};

struct Barodesy : Plasti_rule {			
	double ficr, rparam, N, lambda, kappa;
	Barodesy(Hypo_mode istr) : Plasti_rule(hypoplasticity, istr){}; 
	virtual void inic_hisv(); 
	virtual void give_LN() {/*nothing done*/};
	virtual void give_barodesy_Rab(double Dstretch[3][3]);
};

struct Hypoplasti_clay_unsat_expansive : Hypoplasti_clay_unsat  {
	double eM0, kappam, mmparam, lambdapactive, aer,lambdafact,smstar,emstar,re;
	int scanning;
	Hypoplasti_clay_unsat_expansive(Hypo_mode istr, int add=3) : Hypoplasti_clay_unsat(istr, add){}; 
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void inic_hisv(); 
	virtual void make_sig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double tensor_unsat[MDIM*MDIM], double &scalar_unsat);
	virtual void make_dsig_unsat(double sig[MDIM*MDIM], double ua, double uw, 
		double dsig[MDIM*MDIM], double dua, double duw, 
		double dtensor_unsat[MDIM*MDIM], double &dscalar_unsat);
	virtual void calc_depm_suction(double delta_sig[9], double d_ua, double d_uw,double delta_epadd[9],double &factor);
	virtual bool unsat_stress_varWRC_correction(double dep_Meef[9], double delta_temper, double &correction);
	virtual void calc_N_lambda_dpeds(double &Nuse, double &lambda_use, double &dpeds_divpe, double &dpedT_divpe);
};

struct Hypoplasti_clay_unsat_Sr : Hypoplasti_clay_unsat  {
	double e0;
	Hypoplasti_clay_unsat_Sr(Hypo_mode istr, int add=3) : Hypoplasti_clay_unsat(istr, add){}; 
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void inic_hisv(); 
	virtual void calc_sairentry(bool init); 
};

struct Hypoplasti_unsat_thermo_Sr : Hypoplasti_clay_unsat  {
	double e0, nTparam, lTparam, Trparam;
	Hypoplasti_unsat_thermo_Sr(Hypo_mode istr, int add=3) : Hypoplasti_clay_unsat(istr, add){}; 
	virtual void calc_N_lambda_dpeds(double &Nuse, double &lambda_use, double &dpeds_divpe, double &dpedT_divpe);
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void inic_hisv();
	virtual void calc_sairentry(bool init); 
	virtual void calc_depm_thermal(double delta_temperature,double delta_ep_thermal[9],double &factor);
};

struct Hypoplasti_unsat_thermo : Hypoplasti_clay_unsat  {
	double Trparam, nTparam, lTparam;
	Hypoplasti_unsat_thermo(Hypo_mode istr, int add=3) : Hypoplasti_clay_unsat(istr, add){}; 
	virtual void calc_N_lambda_dpeds(double &Nuse, double &lambda_use, double &dpeds_divpe, double &dpedT_divpe);
	virtual void update_hisv(double d_ept[9], double d_other_hisv[2], double delta_epadd[9]);
	virtual void inic_hisv();
	virtual void calc_sairentry(bool init); 
	virtual void calc_depm_thermal(double delta_temperature,double delta_ep_thermal[9],double &factor);
};

struct Hypoplasti_clay_unsat_multilin : Hypoplasti_clay_unsat {
	vector<double> Nltriplets;
	Hypoplasti_clay_unsat_multilin(Hypo_mode istr, int add=2) : Hypoplasti_clay_unsat(istr, add){}; 
	virtual void calc_N_lambda_dpeds(double &Nuse, double &lambda_use, double &dpeds_divpe, double &dpedT_divpe);
	virtual void inic_hisv(); 
};

struct Cloe : Plasti_rule {						
	double cloe_param[37];
	bool clay_mode;
	Cloe(Hypo_mode istr, bool cm) : Plasti_rule(hypoplasticity, istr, 29), clay_mode(cm) {}; 
	virtual void inic_hisv(); 
	virtual void give_LN();
};

struct Umat : General_model {
	Umat() : General_model(-1,-1,6,6) {}; //nparms and nstatev will be overwritten after the first umat call

        //Main interface routine
        virtual void soil_model(double strain_gen[], double stress_gen[], double qstatev[], 
        double dstrain_gen[], double dstress_gen[], double dqstatev[], double *DDtan_gen, 
        double parms[], int flag);
};

struct Umatunsat : Plasti_rule {
	Umatunsat() : Plasti_rule(umatunsatmodel) {};
	void inic_hisv(); 
};

struct Generalmodel_plastirule : Plasti_rule {
	Generalmodel_plastirule() : Plasti_rule(generalmodel) {};
};

extern "C" {void cloe_( double&, double&, double&, double& );}
extern "C" {void cloeclay_( double&, double&, double&, double& );}
extern "C" {void umat_(double& , double& , double& ,  double &, double &, double &, double &, double &, double &, double &, double &, double &,
	  double &, double &, double &, double &, double &, double &, char &,  int &,  int &, int &, int &, double &, int &, double &, double &, double &,
	  double &, double &, double &, int &, int &,  int &,  int &,  int &, int &);}
extern "C" {void umatunsat_(double& , double& , double& ,  double &, double &, double &, double &, double &, double &, double &, double &, double &,
	  double &, double &, double &, double &, double &, double &, char &,  int &,  int &, int &, int &, double &, int &, double &, double &, double &,
	  double &, double &, double &, int &, int &,  int &,  int &,  int &, int &, double &);}

/*******************************VYPOCET*******************************************/

struct Count {			
	string output_file;
	Elasti_rule* elasti;
	Plasti_rule* plasti;
        General_model* general;
	vector<double> triaxdata;			
	double step, max_value, tan_dir, stepsig;
	Control_mode control_mode;
	Test_type threeDtest;
	bool mode3D;
	bool stresscontrol;
	int start_data;		    //First control parameter (normal -> 2; stressdir -> 3)
	int number_of_input_values; //Number of parameters to controll one step of the calculation
				    //(constdir_pq -> 2; normal -> 1)
	double Kw;
	Count(Elasti_rule* a, Plasti_rule* b, General_model* g, Control_mode str=no_spec, bool m3D=false, int st=2)
			 : elasti(a), plasti(b), general(g), control_mode(str), mode3D(m3D), start_data(st) {
			tan_dir=0;
			threeDtest=not_spec;
			stresscontrol=false;
			number_of_input_values=1; 
	};
	virtual void count_elasti(bool trial);		
	virtual void count_plasti();	
	virtual void count_hypoplasti ();
	virtual void count_umat ();
	virtual void count_umatunsat ();
        virtual void count_general();
        virtual void update_general();
	virtual bool controll_plasti();
	virtual void count_steps();
	virtual void calc_results();
	virtual void initialize();
	virtual void init_note() {};
	virtual void count() {count_main();};
	virtual void count_main();
	virtual void init_general();
	virtual void give_dep(double matrix[MDIM][MDIM][MDIM][MDIM], double delta_ep[MDIM*MDIM]) {};
	virtual void give_dep_hypoplasti (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM]);	
	virtual void give_dep_hypoplasti_DdepL (double hypo_L[MDIM][MDIM][MDIM][MDIM], double hypo_L_unl[MDIM][MDIM][MDIM][MDIM],
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM]);	
	virtual bool give_dep_straincontrol (double delta_ep[MDIM*MDIM]);
	virtual void give_dep_hypoplasti_stress (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double d_sig[MDIM*MDIM], double delta_ep[MDIM*MDIM]);	
	virtual void give_dsig(double d_sig[MDIM*MDIM]);
	virtual void give_dsig_unsat(double d_sig[MDIM*MDIM],double &d_ua, double &d_uw) {};
	virtual void give_dep_hypoplasti_stress_unsat (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double d_tensor_unsat[MDIM*MDIM], double d_scalar_unsat, 
		double delta_ep[MDIM*MDIM]) {};	
	virtual void give_Kw(double &Kw) {Kw=0;};
        virtual void give_generalmod_ESn(double *Ematrix, double *Smatrix, double nuvect[], int ngstrain, int ngstress);
};

struct Straincontrol_wholetensor : Count {	
	Straincontrol_wholetensor(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Count(an, bn, gn) {};
	virtual void init_note();
	virtual bool give_dep_straincontrol (double delta_ep[MDIM*MDIM]);	
};

struct Strain_cycle : Count {	
	Strain_cycle(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Count(an, bn, gn) {};
	virtual void init_note();
	virtual bool give_dep_straincontrol (double delta_ep[MDIM*MDIM]);	
};

struct Count_unsat : Count {
	double stage_init_epp[9];
	Count_unsat(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=2)
		 : Count(an, bn, gn, strdr, false, st) {array_set(stage_init_epp,0,9);};
	virtual bool controll_plasti();
	virtual void count_plasti();	
	virtual void count_elasti(bool trial);		
	virtual void give_dsig_unsat(double d_sig[MDIM*MDIM],double &d_ua, double &d_uw) {};
	virtual void count_hypoplasti ();
	virtual void give_dep_hypoplasti_stress_unsat (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double d_tensor_unsat[MDIM*MDIM], double d_scalar_unsat, 
		double delta_ep[MDIM*MDIM]);	
	virtual void give_dep_hypoplasti_unsat(double hypo_L[MDIM][MDIM][MDIM][MDIM],
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM], double &dua, double &duw, double dsignet[9]);
 	virtual void init_note();
};

struct Count_temper : Count {	
	Count_temper(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=2)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void give_dsig_temper(double d_sig[MDIM*MDIM],double &d_ua, double &d_uw, double &delta_temper) {};
	virtual void count_hypoplasti ();
	virtual void give_dep_hypoplasti_stress_temper (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double d_tensor_unsat[MDIM*MDIM], double d_scalar_unsat, 
		double delta_ep_eff[MDIM*MDIM], double delta_temper);
};

struct Count_temper_constwc_sat : Count {	
	Count_temper_constwc_sat(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=2)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void give_dsig_temper(double d_sig[MDIM*MDIM], double &delta_temper) {};
	virtual void count_hypoplasti ();
	virtual void give_dep_hypoplasti_stress_temper (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double delta_sig[MDIM*MDIM], double &d_scalar_unsat, 
		double delta_ep_eff[MDIM*MDIM], double delta_temper);
};

struct Count_temper_constwc_unsat : Count {	
	Count_temper_constwc_unsat(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=2)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void give_dsig_temper(double d_sig[MDIM*MDIM], double &delta_temper) {};
	virtual void count_hypoplasti ();
	virtual void give_dep_hypoplasti_stress_temper (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double delta_sig[MDIM*MDIM], double &d_scalar_unsat, 
		double delta_ep_eff[MDIM*MDIM], double delta_temper);
};

/*******************************Const dir. in the stress space tests***********************/

struct Triax_stressdir : Count {	
	Triax_stressdir(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=3)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void give_dep(double matrix[MDIM][MDIM][MDIM][MDIM], double delta_ep[MDIM*MDIM]);
	virtual void give_dep_hypoplasti (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM]);	
	virtual void give_dep_hypoplasti_DdepL (double hypo_L[MDIM][MDIM][MDIM][MDIM], double hypo_L_unl[MDIM][MDIM][MDIM][MDIM],
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM]);	
	virtual void give_dep_hypoplasti_stress (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double d_sig[MDIM*MDIM], double delta_ep[MDIM*MDIM]);	
	virtual void give_dsig(double d_sig[MDIM*MDIM]);
	virtual void init_note();
	virtual void give_dir(double &dir_deg);
};

struct Triax_constdir_pq : Triax_stressdir {	
	Triax_constdir_pq(Elasti_rule* anf, Plasti_rule* bnf, General_model* gnf, Control_mode strdrf=strain_auto) : 
		Triax_stressdir(anf, bnf, gnf, strdrf, false, 2) {};
	virtual void give_dir(double &dir_deg);
};

struct Triax_constp : Triax_stressdir {	
	Triax_constp(Elasti_rule* anf, Plasti_rule* bnf, General_model* gnf, Control_mode strdrf=strain_auto) : 
		Triax_stressdir(anf, bnf, gnf, strdrf, false, 2) {};
	virtual void give_dir(double &dir_deg) {dir_deg=90;};
};

struct Triax_drained : Triax_stressdir {	
	Triax_drained(Elasti_rule* anf, Plasti_rule* bnf, General_model* gnf, Control_mode strdrf=strain_auto) : 
		Triax_stressdir(anf, bnf, gnf, strdrf, false, 2) {};
	virtual void give_dir(double &dir_deg) {dir_deg=atan(3.)*180/PI;};
        virtual void give_generalmod_ESn(double *Ematrix, double *Smatrix, double nuvect[], int ngstrain, int ngstress);
};

struct Triax_undrained_Kw : Triax_stressdir {	
	Triax_undrained_Kw(Elasti_rule* anf, Plasti_rule* bnf, General_model* gnf, Control_mode strdrf=strain_auto) : 
		Triax_stressdir(anf, bnf, gnf, strdrf, false, 2) {};
	virtual void give_dir(double &dir_deg) {dir_deg=atan(3.)*180/PI;};
	virtual void give_Kw(double &Kw) {Kw=1.e9;};
};

struct Triax_constax : Triax_stressdir {	
	Triax_constax(Elasti_rule* anf, Plasti_rule* bnf, General_model* gnf, Control_mode strdrf=strain_auto) : 
		Triax_stressdir(anf, bnf, gnf, strdrf, false, 2) {};
	virtual void give_dir(double &dir_deg) {dir_deg=atan(-1.5)*180/PI;};
};

struct Triax_constq : Triax_stressdir {	
	Triax_constq(Elasti_rule* anf, Plasti_rule* bnf, General_model* gnf, Control_mode strdrf=strain_auto) : 
		Triax_stressdir(anf, bnf, gnf, strdrf, false, 2) {};
	virtual void give_dir(double &dir_deg) {dir_deg=0;};
        virtual void give_generalmod_ESn(double *Ematrix, double *Smatrix, double nuvect[], int ngstrain, int ngstress);
};

struct Simple_shear : Count {	
	Simple_shear(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=2)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void give_dep(double matrix[MDIM][MDIM][MDIM][MDIM], double delta_ep[MDIM*MDIM]);
	virtual void give_dep_hypoplasti (double hypo_L[MDIM][MDIM][MDIM][MDIM], 
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM]);	
	virtual void init_note();
};

/*********************************************************************************************/

struct Strain_resp_envelope : Triax_stressdir {	
	double init_sig[MDIM*MDIM];
	double init_epe[MDIM*MDIM];
	double init_epp[MDIM*MDIM];
	double Dstep[MDIM*MDIM];
	vector<double> init_hisv;
	Strain_resp_envelope(Elasti_rule* anf, Plasti_rule* bnf, General_model* gnf, Control_mode strdrf=strain_auto) : 
		Triax_stressdir(anf, bnf, gnf, strdrf, false, 2) {};
	virtual void count();
	virtual void init_note() {};
};

struct Camclay_sbs : Count {	
	double init_sig[MDIM*MDIM];
	double init_e, init_pc;
	Camclay_sbs(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=3)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void count();
	virtual bool loglog() {return(false);};
};

struct Camclay_loglog_sbs : Camclay_sbs {	
	Camclay_loglog_sbs(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=3)
		 : Camclay_sbs(an, bn, gn, strdr, false, st) {};
	bool loglog() {return(true);};
};

struct Hypo_sbs : Count {	
	double init_sig[MDIM*MDIM];
	double init_e, init_ua, init_uw;
	Hypo_sbs(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=3)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void count();
};

struct Hypo_sbs_soft : Count {	
	double init_sig[MDIM*MDIM];
	double init_e, init_s;
	Hypo_sbs_soft(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=3)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void count();
};

struct Hypo_vW_sbs : Count {	
	double init_sig[MDIM*MDIM];
	double init_e;
	Hypo_vW_sbs(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr, bool k=false, int st=3)
		 : Count(an, bn, gn, strdr, false, st) {};
	virtual void count();
};

struct Triax_constp3D : Count {	
	Triax_constp3D(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode ctr) :
		 Count(an, bn, gn, ctr, true, 3) {};
	virtual void give_dep(double matrix[MDIM][MDIM][MDIM][MDIM], double delta_ep[MDIM*MDIM]);
	virtual void give_dsig(double d_sig[MDIM*MDIM]);
	virtual void init_note();
};

struct Triax_cpcirc3D : Triax_constp3D {	
	Triax_cpcirc3D(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Triax_constp3D(an, bn, gn, stress) {};
	virtual void init_note();
	virtual void count();
};

struct Triax_straindir : Count {	
	Triax_straindir(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr) : Count(an, bn, gn, strdr, false, 3) {};
	virtual bool give_dep_straincontrol (double delta_ep[MDIM*MDIM]);	
	virtual bool psidir() {return(false);};
};

struct Triax_straindir_psidir : Triax_straindir {	
	Triax_straindir_psidir(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdr) : Triax_straindir(an, bn, gn, strdr) {};
	virtual bool psidir() {return(true);};
};

struct Triax_undrained : Count {	
	Triax_undrained(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Count(an, bn, gn) {};
	virtual bool give_dep_straincontrol (double delta_ep[MDIM*MDIM]);
        virtual void give_generalmod_ESn(double *Ematrix, double *Smatrix, double nuvect[], int ngstrain, int ngstress);
};

struct Simple_shear_undr : Count {	
	Simple_shear_undr(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Count(an, bn, gn) {};
	virtual bool give_dep_straincontrol (double delta_ep[MDIM*MDIM]);	
	virtual void init_note();
};

struct Triax_K0 : Count {	
	Triax_K0(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Count(an, bn, gn) {};
	virtual bool give_dep_straincontrol (double delta_ep[MDIM*MDIM]);
        virtual void give_generalmod_ESn(double *Ematrix, double *Smatrix, double nuvect[], int ngstrain, int ngstress);
};	

struct Triax_K0_unsat : Count_unsat {	
	Triax_K0_unsat(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdrf=strain_ax) : Count_unsat(an, bn, gn, strdrf) {};
	virtual void give_dep_hypoplasti_unsat(double hypo_L[MDIM][MDIM][MDIM][MDIM],
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM], double &dua, double &duw, double dsignet[9]);
};	

struct Triax_unsat_swellingpressure : Count_unsat {	
	Triax_unsat_swellingpressure(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdrf=strain_ax) : Count_unsat(an, bn, gn, strdrf) {};
	virtual void give_dep_hypoplasti_unsat(double hypo_L[MDIM][MDIM][MDIM][MDIM],
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM], double &dua, double &duw, double dsignet[9]);
};	

struct Triax_constwc_unsat : Count_unsat {	
	Triax_constwc_unsat(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdrf=strain_ax) : Count_unsat(an, bn, gn, strdrf) {};
	virtual void give_dep_hypoplasti_unsat(double hypo_L[MDIM][MDIM][MDIM][MDIM],
		double hypo_N[MDIM][MDIM], double delta_ep[MDIM*MDIM], double &dua, double &duw, double dsignet[9]);
};	

struct Triax_unsat_constsignet : Count_unsat {	
	Triax_unsat_constsignet(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdrf=stress) : Count_unsat(an, bn, gn, strdrf) {};
	virtual void give_dsig_unsat(double d_sig[MDIM*MDIM],double &d_ua, double &d_uw);
        virtual void give_generalmod_ESn(double *Ematrix, double *Smatrix, double nuvect[], int ngstrain, int ngstress);
};

struct Triax_unsat_constqnet : Count_unsat {	
	Triax_unsat_constqnet(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdrf=stress) : Count_unsat(an, bn, gn, strdrf, false, 3) {};
	virtual void give_dsig_unsat(double d_sig[MDIM*MDIM],double &d_ua, double &d_uw);
};

struct Triax_temper_constsig : Count_temper {	
	Triax_temper_constsig(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdrf=stress) : Count_temper(an, bn, gn, strdrf) {};
	virtual void give_dsig_temper(double d_sig[MDIM*MDIM],double &d_ua, double &d_uw, double &delta_temper);
        virtual void give_generalmod_ESn(double *Ematrix, double *Smatrix, double nuvect[], int ngstrain, int ngstress);
};

struct Triax_temper_constsig_constwc_sat : Count_temper_constwc_sat {	
	Triax_temper_constsig_constwc_sat(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdrf=stress) : Count_temper_constwc_sat(an, bn, gn, strdrf) {};
	virtual void give_dsig_temper(double d_sig[MDIM*MDIM], double &delta_temper);
};

struct Triax_temper_constsig_constwc_unsat : Count_temper_constwc_unsat {	
	Triax_temper_constsig_constwc_unsat(Elasti_rule* an, Plasti_rule* bn, General_model* gn, Control_mode strdrf=stress) : Count_temper_constwc_unsat(an, bn, gn, strdrf) {};
	virtual void give_dsig_temper(double d_sig[MDIM*MDIM], double &delta_temper);
};

struct Surface : Count {	
	int yield;
	Surface(Elasti_rule* an, Plasti_rule* bn, General_model* gn, int y) : Count(an, bn, gn), yield(y) {};
	virtual void count();
	void count_kinemat();
	void count_hypo_sbs();
	void count_isotr();
	void count_3D();
};

struct Resp_envelope : Count {	
	double init_sig[MDIM*MDIM];
	double init_epe[MDIM*MDIM];
	double init_epp[MDIM*MDIM];
	double Dstep[MDIM*MDIM];
	vector<double> init_hisv;
	int resp_env_type;
	Resp_envelope(Elasti_rule* an, Plasti_rule* bn, General_model* gn, int t) : Count(an, bn, gn) {
		resp_env_type=t;
	};
	virtual void count();
	virtual bool give_dep_straincontrol (double delta_ep[MDIM*MDIM]);	
};

struct Init_pq : Count {	
	Init_pq(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Count(an, bn, gn) {};
	virtual void count();
	virtual void init_uauw() {};
};

struct Init_pq_unsat : Init_pq {	
	Init_pq_unsat(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Init_pq(an, bn, gn) {};
	virtual void init_uauw();
};

struct Init_pq_unsat_temper : Init_pq {	
	Init_pq_unsat_temper(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Init_pq(an, bn, gn) {};
	virtual void init_uauw();
};

struct Init_stresstensor : Count {	
	Init_stresstensor(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Count(an, bn, gn) {};
	virtual void count();
};

struct Rezero_strains : Count {	
	Rezero_strains(Elasti_rule* an, Plasti_rule* bn, General_model* gn) : Count(an, bn, gn) {};
	virtual void count();
};

/*************************Input***********************************/

enum Num_el_model {
	barcelona_basic_elasti,
	camclay_elasti,
	camclay_loglog_elasti,
	camclay_loglog_presdepG_elasti,
	camclay_poisson_elasti,
	gklinp_elasti,
	hypoelasti,
	severntrent_elasti,
	smallstrain_elasti,
	tskh_elasti,
	young,
	young_power,
	young_strainstress,
	young_tris
};

enum Num_pl_model {
	barcelona_basic_plasti,
	barodesy,
	camclay_plasti,
	camclay_soft_plasti,
	cloe,
	cloe_clay,
	druckprag,
	generalmodel_plastirule,
	hypoplasti_camclay,
	hypoplasti_clay,
	hypoplasti_clay_anisot,
	hypoplasti_clay_explicit,
	hypoplasti_clay_soft,
	hypoplasti_clay_unsat,
	hypoplasti_clay_unsat_anisot,
	hypoplasti_clay_unsat_expansive,
	hypoplasti_clay_unsat_multilin,
	hypoplasti_clay_unsat_Sr,
	hypoplasti_clay_unsat_Sr_anisot,
	hypoplasti_huang,
	hypoplasti_lowangles,
	hypoplasti_unsat_thermo,
	hypoplasti_unsat_thermo_Sr,
	hypoplasti_vW,
	hypoplasti_vW_soft,
	lagpotts_camclay_plasti,	
	lagpotts_cem,
	mohrcoul,
	mohrcoul_softening,
	severntrent_plasti,
	tskh_general_plasti,
	tskh_plasti,
	umatunsat
};

enum Num_generalmodel {
	hypoplasti_unsat_expansive_thermal,
	umat
};

enum Num_lab_model {
	all_surfaces,
	all_surfaces_3D,
	all_surfaces_3Dgp,
	camclay_sbs,
	camclay_loglog_sbs,
	hypo_sbs,
	hypo_sbs_soft,
	hypo_vW_sbs,
	init_pq,
	init_pq_unsat,
	init_pq_unsat_temper,
	init_stresstensor,
	pp_surface,
	resp_envelope,
	resp_envelope_incr,
	rezero_strains,
	simple_shear,
	simple_shear_undr,
	strain_cycle,
	strain_resp_envelope,
	strain_resp_envelope_strain,
	straincontrol_wholetensor,
	triax_constax,
	triax_constdir_pq,
	triax_constdir_pq_ax,
	triax_constdir_pq_rad,
	triax_constdir_pq_stress,
	triax_constp,
	triax_constp3D,
	triax_constp3D_stress,
	triax_constq,
	triax_constwc_unsat,
	triax_cpcirc3D,
	triax_drained,
	triax_K0,
	triax_K0_unsat,
	triax_straindir,
	triax_straindir_psidir,
	triax_stressdir,
	triax_stressdir_ax,
	triax_stressdir_rad,
	triax_stressdir_stress,
	triax_temper_constsig,
	triax_temper_constsig_constwc_sat,
	triax_temper_constsig_constwc_unsat,
	triax_undrained,
	triax_undrained_Kw,	
	triax_unsat_constqnet,
	triax_unsat_constsignet,
	triax_unsat_swellingpressure,
	yield_surface,
	
	next_step,	//changes step, no model; Always at the end!!!
	change_step	//changes step, no model
};

struct Input {
	vector<Elasti_rule*> vector_elasti;
	vector<Plasti_rule*> vector_plasti;
        vector<General_model*> vector_generalmod;
	vector<Count*> vector_lab;
	vector<double> totaldata;
	char input_file[MCHAR], output_file[MCHAR];
	int num_results, num_data, actual_stage;
	vector<Num_lab_model> test_for_stage;
	vector<Condition> cond_for_stage;
	vector<int> stage_order, num_stagedata, noresults_stg, results_stg;
	vector<int> stage_for_condition;
	vector<double> total_triaxdata;
	Tskh_mode tskh_mode;
	bool hypo_presdepe, external_plasti_data;
	Hypo_mode hypo_mode;
	bool tris;
	Input() {
		num_results=num_data=actual_stage=0;
		tris=false;
		tskh_mode=tskh;
	}; 
	void main(char* inp, char* out);
	void make_outputfile();
	Num_el_model num_elasti;
	Num_pl_model num_plasti;
        Num_generalmodel num_generalmod;
	Num_lab_model num_lab;
};
