#include <string.h>
#include <stdlib.h>
#include "bocon.h"
#include "parser.h"
#include "globprep.h"


/**
  Constructor initializes data members
  Parameters :
    none
  Returns :
    nothing
*/
bocon::bocon()
{
  nn = 0L;
  ndir = 0L;
  dir = NULL;
  con = NULL;
  expr = NULL;
  nspd = NULL;
  ndpd = NULL;
}



/**
  Method reads data from the text file specified by parameter in, data are
    in preprocessor format
  Parameters :
    @param in    - pointer to the opened text file, where the data will be read
    @param ndofn - number of dofs in given node with prescribed condition
  Returns :
    @retval 0 - on success
    @retval 1 - number of boundary conditions is not in range <1, ndofn>
    @retval 2 - direction of a boundary condition is not in range <1, ndofn>
    @retval 3 - load case of a boundary condition is not in range <1, Nlc>
    @retval 4 - subload case of a boundary condition is not in range <1, Nslc>
    @retval 5 - error in parsing string for dynamic boundary condition
*/
long bocon::read(XFILE *in, long ndofn)
{
  long nbc, td, tlc, tslc, tnlc, i;
  double tc;
  Parser par;
  Equation *eq;

  // deallocate eventually used memory
  if (dir)
    delete [] dir;
  if (con)
    delete [] con;
  if (nspd)
    delete [] nspd;
  if (ndpd)
    delete [] ndpd;
  if (expr)
  {
    for (i = 0; i < ndir; i++)
      delete [] expr[i];
    delete [] expr;
  }

  tnlc = Nlc;
  if (Tnslc)
    tnlc = Tnslc;

  ndir = ndofn;
  xfscanf (in, "%k%ld", "num_bc", &nbc);
  // reading number of conditions
  if ((nbc < 1) || (nbc > ndir))
  // which should be in range <1,ndofn>
    return(1);
  // allocating memory for arrays
  dir = new long[ndir];
  for (i = 0; i < ndir; dir[i] = -1, i++);
  // if dir[i] == -1 there is no codition in the i direction
  nspd = new long[ndir];
  memset(nspd, 0, sizeof(*nspd)*ndir);
  ndpd = new long[ndir];
  memset(ndpd, 0, sizeof(*ndpd)*ndir);
  con = new double[ndir*tnlc];
  memset(con, 0, sizeof(*con)*ndir*tnlc);
  expr = new char*[ndir*tnlc];
  memset(expr, 0, sizeof(*expr)*ndir*tnlc);

  // loop which reads conditions in particular directions
  for (i = 0; i < nbc; i++)
  {
    xfscanf(in, "%k%ld", "dir", &td);  // reading direction
    xfscanf(in, "%k%le", "cond", &tc); // reading condition
    if ((labs(td) < 1) || (labs(td) > ndir))
    // direction should be in range <1, ndir>
      return(2);
    if (td > 0)
    {
      td--;
      dir[td] = td;
      tlc = 0;
      if (tc != 0.0) 
      // load case id is required for nonzero prescribed displacements
      {
        xfscanf (in, "%k%ld", "lc_id", &tlc);
        if ((tlc < 1) || (tlc > Nlc))
          return (3);
        tlc--;
        if (Tnslc)
        {
          xfscanf(in, "%k%ld", "slc_id", &tslc);
          if ((tslc < 1) || (tslc > Nslc[tlc]))
            return (4);
          tslc--;
          tlc = Nslc_cum[tlc]+tslc;
        }
        nspd[td]++; 
      }
      con[tlc*ndir+td] = tc;
    }
    if (td < 0)
    // if direction number is negative -> dynamic condition
    {
      td = labs(td);
      td--;
      dir[td] = td;
      tlc = 0;
      if (tc != 0.0) 
      // load case id is required for nonzero prescribed displacements
      {
        xfscanf (in, "%k%ld", "lc_id", &tlc);
        if ((tlc < 1) || (tlc > Nlc))
          return (3);
        tlc--;
        if (Tnslc)
        {
          xfscanf(in, "%k%ld", "slc_id", &tslc);
          if ((tslc < 1) || (tslc > Nslc[tlc]))
            return (4);
          tslc--;
          tlc = Nslc_cum[tlc]+tslc;
        }
        expr[tlc*ndir+td] = new char[1025];
        expr[tlc*ndir+td][0] = '\0';
        xfscanf(in, "%1024s", expr[tlc*ndir+td]); // read string with expression
        eq = par.TextToTree(expr[tlc*ndir+td]); // expansion expression in the string
                                    // by parser to binary tree
        if (eq == NULL)
          // expansion failed
          return(5);
        ndpd[td]++; 
      }
      else
        nspd[td]++; 

      con[tlc*ndir+td] = tc;
    }
  }
  return(0);
}



/**
  Method tries to merge two boundary conditions

  @param tbc  - object of boundary condition which will be copied

  @retval Method returns pointer to new object of boundary condition
*/
bocon *bocon::merge(bocon *tbc)
{
  bocon *ret = new bocon;
  long i, k, l, nlcid;

  ret->nn = nn;
  ret->ndir = ndir;
  nlcid = Nlc;
  if (Tnslc)
    nlcid = Tnslc;

  ret->dir = new long[ndir];
  for (i = 0; i < ndir; ret->dir[i] = -1, i++);
  ret->nspd = new long[ndir];
  memset(ret->nspd, 0, sizeof(*ret->nspd)*ndir);
  ret->ndpd = new long[ndir];
  memset(ret->ndpd, 0, sizeof(*ret->ndpd)*ndir);
  ret->con = new double[ndir*nlcid];
  memset(ret->con, 0, sizeof(*ret->con)*ndir*nlcid);
  ret->expr = new char*[ndir*nlcid];
  memset(ret->expr, 0, sizeof(*ret->expr)*ndir*nlcid);

  for (k=0; k < nlcid; k++)
  {
    for (i=0; i<ndir; i++)
    {
      // compare BC direction indicators
      if (dir[i] == tbc->dir[i])
      {      
        // both BC have prescribed condition in this direction
        ret->dir[i] = dir[i];
        // compare BC condition type
        if (con[ndir*k+i] == tbc->con[ndir*k+i])
        {
          // values of conditions are the same
          ret->con[ndir*k+i] = con[ndir*k+i];
          if ((tbc->expr[ndir*k+i] == NULL) && (expr[ndir*k+i] == NULL))
          {
            // no BC for dynamics 
            ret->expr[ndir*k+i] = NULL;
            if (con[ndir*k+i] != 0.0)
              ret->nspd[i]++;
            continue;
          }
          else
          {
            if (strcmp(expr[ndir*k+i], tbc->expr[ndir*k+i]) == 0)
            {
              // time functions for dynamic have to be indentical
              l = strlen(expr[ndir*k+i]);
              ret->expr[ndir*k+i] = new char[l+1];
              memcpy(ret->expr[ndir*k+i], expr[ndir*k+i], sizeof(*expr[ndir*k+i])*(l+1));
              ret->ndpd[i]++;
              continue;
            }
            else
            // conflict in merged conditions - expressions are different
            {
              return NULL;
            }
          }
        }
        else
        // conflict in the merged conditions - 
        // different values of boundary conditions in the same direction cannot be merged
        {
          return NULL;
        }
      }
      else
      {
        // only one of the BC has been assigned to this direction
        if (dir[i] != -1)
        {
          // copy dir of this BC
          ret->dir[i] = dir[i];
          ret->con[ndir*k+i] = con[ndir*k+i];
          if (expr[ndir*k+i] == NULL)
          {
            ret->expr[ndir*k+i] = NULL;
            if (con[ndir*k+i] != 0.0)
              ret->nspd[i]++;
          }
          else
          {
            l = strlen(expr[ndir*k+i]);
            ret->expr[ndir*k+i] = new char[l+1];
            memcpy(ret->expr[ndir*k+i], expr[ndir*k+i], sizeof(*expr[ndir*k+i])*(l+1));
            ret->ndpd[i]++;
            continue;
          }
        }
        else
        {
          // copy dir of tbc
          ret->dir[i] = tbc->dir[i];
          ret->con[ndir*k+i] = tbc->con[ndir*k+i];
          if (tbc->expr[ndir*k+i] == NULL)
          {
            ret->expr[ndir*k+i] = NULL;
            if (tbc->con[ndir*k+i] != 0.0)
              ret->nspd[i]++;
          }
          else
          {
            l = strlen(tbc->expr[ndir*k+i]);
            ret->expr[ndir*k+i] = new char[l+1];
            memcpy(ret->expr[ndir*k+i], tbc->expr[ndir*k+i], sizeof(*tbc->expr[ndir*k+i])*(l+1));
            ret->ndpd[i]++;
            continue;
          }
        }        
      }
    }
  }
  return ret;
}



/**
  Method creates copy of given object 

  Returns:
  Method returns pointer to newly alocated object of boundary condition.
*/
bocon *bocon::copy(void)
{
  bocon *ret = new bocon;
  long i, k, l, nlcid;

  ret->nn = nn;
  ret->ndir = ndir;
  nlcid = Nlc;
  if (Tnslc)
    nlcid = Tnslc;

  ret->dir = new long[ndir];
  memcpy(ret->dir, dir, sizeof(*dir)*ndir);
  ret->nspd = new long[ndir];
  memcpy(ret->nspd, nspd, sizeof(*nspd)*ndir);
  ret->ndpd = new long[ndir];
  memcpy(ret->ndpd, ndpd, sizeof(*ndpd)*ndir);
  ret->con = new double[ndir*nlcid];
  memcpy(ret->con, con, sizeof(*con)*ndir*nlcid);
  ret->expr = new char*[ndir*nlcid];
  memset(ret->expr, 0, sizeof(*ret->expr)*ndir*nlcid);
  for (k=0; k < nlcid; k++)
  {
    for(i=0; i<ndir; i++)
    {
      if (expr[ndir*k+i])
      {
        l = strlen(expr[ndir*k+i]);
        ret->expr[ndir*k+i] = new char[l+1];
        memcpy(ret->expr[ndir*k+i], expr[ndir*k+i], sizeof(*expr[ndir*k+i])*(l+1));
      }
    }
  }
  return ret;
}



bocon::~bocon()
{
  long tnlc = Nlc;
  if (Tnslc)
    tnlc = Tnslc;

  delete [] dir;
  delete [] con;
  delete [] nspd; 
  delete [] ndpd; 
  if (expr)
  {
    for (long i = 0; i < tnlc*ndir; i++)
      delete [] expr[i];
    delete [] expr;
  }
}
