#include "creep_effym.h"
#include "matrix.h"
#include "vector.h"
#include "elastisomat.h"
#include "global.h"
#include "intpoints.h"
#include "vecttens.h"
#include <math.h>

/**
  This constructor inializes attributes to zero values.
*/
creep_effym::creep_effym (void)
{

}

/**
  This destructor is only for the formal purposes.
*/
creep_effym::~creep_effym (void)
{

}



/**
  Function computes material stiffnes %matrix.

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number
  @param im - material index
  @param ido - index of internal variables for given material in the ipp eqother array


*/
void creep_effym::matstiff (matrix &d,long ipp,long im, long ido)
{
  long idem = Mm->ip[ipp].gemid();
  long ncomp = Mm->ip[ipp].ncompstr;
  Mm->matstiff(d, ipp, idem, ido+ncomp);
}



/**
  Function computes correct stresses in the integration point and stores
  them into ip stress array. This material is reasonable only in combination with 
  ELASTTIME material, which deals with effective Young modulus.
  
  @param ipp - integration point pointer
  @param im  - index of the material in the ipp tm and idm arrays. The standard value is zero.
  @param ido - index of internal variables for given material in the ipp other array

  7.2008 TKo
*/
void creep_effym::nlstresses (long ipp, long im, long ido)
{
  long i, idem;
  long ncomp = Mm->ip[ipp].ncompstr;
  double e, e0, nu;
  vector epsn(ncomp),eps_cr(ncomp),sigma(ncomp);
  matrix d(ncomp,ncomp);


  if (Mm->ip[ipp].ssst == planestress)
  {
    idem = Mm->ip[ipp].gemid();
    nu = Mm->eliso[Mm->ip[ipp].idm[idem]].nu;
    Mm->ip[ipp].strain[3] = -nu / (1.0 - nu) * (Mm->ip[ipp].strain[0]+Mm->ip[ipp].strain[1]);
  }

  e = Mm->give_actual_ym(ipp, im, ido);   // effective Young modulus
  e0 = Mm->give_initial_ym(ipp, im, ido); // initial Young modulus
  //  initial values of total strains
  for (i=0;i<ncomp;i++){
    epsn[i] = Mm->ip[ipp].strain[i];
  }
  Mm->elmatstiff (d,ipp);
  
  //  computation of stresses and "irreversible creep strains" from the effectiv Young modulus
  for (i=0;i<ncomp;i++)
  {
   eps_cr[i] = (e0-e)/e0*epsn[i];
   Mm->ip[ipp].other[ido+i] = eps_cr[i];
  }
  subv(epsn, eps_cr, eps_cr);
  mxv(d, eps_cr, sigma);   
  //  new stress data storage
  for (i=0;i<ncomp;i++){
    Mm->ip[ipp].stress[i]=sigma[i];
  }
}


/**
  This function updates values in the other array reached in the previous equlibrium state to
  values reached in the new actual equilibrium state.

  @param ipp - integration point number in the mechmat ip array.
  @param im - material index
  @param ido - index of internal variables for given material in the ipp other array

*/
void creep_effym::updateval (long ipp, long im, long ido)
{
  long i,n = Mm->givencompeqother(ipp,im);

  for (i=0;i<n;i++){
    Mm->ip[ipp].eqother[ido+i]=Mm->ip[ipp].other[ido+i];
  }
}



/**
  This function updates values in the other array reached in the previous equlibrium state to
  values reached in the new actual equilibrium state.

  @param ipp - integration point number in the mechmat ip array.
  @param im - material index
  @param ido - index of internal variables for given material in the ipp other array
  @param epscr - vector of irreversible strains
*/
void creep_effym::giveirrstrains (long ipp, long im, long ido, vector &epscr)
{
  long i,n = Mm->givencompeqother(ipp,im);

  for (i=0;i<n;i++){
    epscr[i]=Mm->ip[ipp].eqother[ido+i];
  }
}



/**
  This function returns the value of tensile strength

  @param ipp - integration point number in the mechmat ip array.
  @param im - material index
  @param ido - index of internal variables for given material in the ipp other array
  
  TKo
*/
double creep_effym::give_actual_ft (long ipp, long im, long ido)
{
  double ft;
  if (im > 0)
    ft = Mm->give_actual_ft(ipp, im+1, ido);
  else
  {
    print_err("creep_effym model does not contain tensile strength", __FILE__, __LINE__, __func__);
    abort();
  }
  return ft;
}
