#include "fixortodam.h"
#include "global.h"
#include "sequent.h"
#include "intpoints.h"
#include "matrix.h"
#include "vecttens.h"


fixortodam::fixortodam (void)
{
  x1t = x2t = x3t = 0.0;
  allocv(3, xeqtf);
  allocv(3, xeqt0);
}



fixortodam::~fixortodam (void)
{
}



/**
  Function reads material parameters from the opened text file.
   
  @param in - pointer to the opened XFILE

  @return The function does not return anything.

  Created by Tomas Koudelka, 11.2014
*/

void fixortodam::read (XFILE *in)
{
  xfscanf(in, "%k %le", "x1t", &x1t);
  xfscanf(in, "%k %le", "x2t", &x2t);
  xfscanf(in, "%k %le", "x3t", &x3t);

  xfscanf(in, "%k %le", "xeq1tf", &xeqtf(0));
  xfscanf(in, "%k %le", "xeq2tf", &xeqtf(1));
  xfscanf(in, "%k %le", "xeq3tf", &xeqtf(2));
}



/**
  The function prints material parameters into the opened text file given
  by the parameter out.

  @param out - pointer to the opened input text file
  
  @return The function does not return anything.

  Created by Tomas Koudelka, 20.1.2015
*/
void fixortodam::print (FILE *out)
{
  fprintf(out, "%le ", x1t);
  fprintf(out, "%le ", x2t);
  fprintf(out, "%le ", x3t);

  fprintf(out, "%le ", xeqtf(0));
  fprintf(out, "%le ", xeqtf(1));
  fprintf(out, "%le ", xeqtf(2));
}



/**
   This function initializes material model data

   @param ipp - integration point number
   @param ido - index of internal variables for given material in the ipp other array

   12/06/2012 TKo
*/
void fixortodam::initval(long ipp, long ido)
{
  matrix d(6,6);

  // It is supposed that elastic stiffness matrix is taken from the elastic orthotropic material 
  // which is given in the local coord. system of the material
  Mm->loc_elmatstiff (d,ipp);

  xeqt0(0) = x1t/d(0,0);
  xeqt0(1) = x2t/d(1,1);
  xeqt0(2) = x3t/d(2,2);
}



void fixortodam::matstiff (matrix &d,long ipp,long ido)
{
  matrix dl(6,6), dg(6,6);
  matrix tmat(3,3);

  switch (Mp->nlman->stmat)
  {
    case initial_stiff:
      // It is supposed that elastic stiffness matrix is taken from the elastic orthotropic material 
      // which is given in the local coord. system of the material
      Mm->loc_elmatstiff (dl,ipp);
      break;
    case tangent_stiff:
      tmatstiff(dl, ipp, ido);
      break;
    default:
      print_err("unknown type of stifness matrix is required.", __FILE__, __LINE__, __func__);
  }

  // evaluate transformation matrix
  Mm->loc_transf_mat(tmat, ipp);

  // transformation to the global coordinate system and conversion 
  // to the reduced form according to stress/straint state
  lg_tens4transf(dg, dl, tmat, Mm->ip[ipp].ssst);
  // convert stiffness matrix(6,6) to reduced form depending on ssst
  tensor4_matrix(d, dg, Mm->ip[ipp].ssst);
}



void fixortodam::tmatstiff (matrix &d,long ipp,long ido)
{
  long i;
  vector omega(3);

  // attained values of damage parameters from the last equilibrium state
  for (i=0; i<3; i++)
    omega(i) = Mm->ip[ipp].eqother[ido+i];

  secstiffmat(d, ipp, omega);
}



/**
  The function returns actual values of equivalent displacements
  in the xeq.

  @param epst - strain tensor in the local coordinate system of the material
  @param xeq  - equivalent displacement %vector (output parameter)

*/
void fixortodam::compute_eqdispl(long ipp, matrix &epst, vector &xeq)
{   
  long i;
  long n = Mm->ip[ipp].ncompstr;
  double l;
  matrix dl(6,6);
  vector eps(n);
  vector sig(n);
  matrix sigt(3,3);

  // It is supposed that elastic stiffness matrix is taken from the elastic orthotropic material 
  // which is given in the local coord. system of the material
  Mm->loc_elmatstiff (dl,ipp);

  long eid = Mm->elip[ipp];
  switch (Mt->give_dimension(eid))
  {
    case 1:
      l = Mt->give_length(eid);
      break;
    case 2:
      l = sqrt(Mt->give_area(eid));
      break;
    case 3:
      l = pow(Mt->give_volume(eid), 1.0/3.0);
      break;
    default:
      print_err("unknown dimension of element", __FILE__, __LINE__, __func__);
  }
  tensor_vector_full(eps, epst, strain);
  mxv(dl,eps,sig);
  vector_tensor(sig,sigt,Mm->ip[ipp].ssst,stress);

  for (i=0; i<3; i++)
  {
    xeq(i) = sigt(i,i)/dl(i,i) * l;
    if (xeq(i) < 0.0)
      xeq(i) = 0.0;
  }
  
  
  
/*  for (i=0; i<3; i++)
    xeq(i)=l*sqrt(pow(epst(i,i),2));  */
}



/**
  The function returns actual values of damage parameters in the particular
  loacl material directions stored in the %vector omega

  @param xeq  - equivalent displacement %vector
  @param omegao - %vector of damage parametrs from the last equilibrium state
  @param omega - %vector of calculated damage parametrs (output parameter)

*/
void fixortodam::compute_dam (vector &xeq, vector &omegao, vector &omega)
{ 
  long i;

  for (i=0; i<3; i++)
  {
    if (xeq(i) >= xeqtf(i))
    {
      omega(i) = 1.0;
      continue;
    }

    if (xeq(i) <= xeqt0(i))
    {
      omega(i) = 0.0;
      continue;
    }
    omega(i)=xeqtf(i)*(xeq(i)-xeqt0(i))/(xeq(i)*(xeqtf(i)-xeqt0(i)));
    if (omega(i)<omegao(i))
      omega(i)=omegao(i);
    if (omega(i)<0)
      omega(i)=0;
  }
}



/**
  The function returns actual values of damage parameters in the particular
  local material directions stored in the %vector omega


  @param d - secant stiffness %matrix  (output parameter)
  @param omega - %vector of damage parameters 
  @param ipp - integration point id

*/
void fixortodam::secstiffmat(matrix &d, long ipp, vector &omega)
{ 
  // It is supposed that elastic stiffness matrix is taken from the elastic orthotropic material 
  // which is given in the local coord. system of the material
  Mm->loc_elmatstiff (d,ipp);

  d(0,0)=pow(1-omega(0),2)*d(0,0);
  d(0,1)=(1-omega(0))*(1-omega(1))*d(0,1);
  d(0,2)=(1-omega(0))*(1-omega(2))*d(0,2);
  d(1,0)=d(0,1);
  d(1,1)=pow(1-omega(1),2)*d(1,1);
  d(1,2)=(1-omega(1))*(1-omega(2))*d(1,2);
  d(2,0)=d(0,2);
  d(2,1)=d(1,2);
  d(2,2)=pow(1-omega(2),2)*d(2,2);
  if (omega(0) == 1.0 && omega(1) == 1.0)
	  d(3, 3) = 0;
  else
	  d(3, 3) = pow((2 * (1 - omega(0))*(1 - omega(1))) / (2 - omega(0) - omega(1)), 2)*d(3,3);
  if (omega(0) == 1.0 && omega(2) == 1.0)
	  d(4, 4) = 0;
  else
	  d(4, 4) = pow((2 * (1 - omega(0))*(1 - omega(2))) / (2 - omega(0) - omega(2)), 2)*d(4,4);
  if (omega(1) == 1.0 && omega(2) == 1.0)
	  d(5, 5) = 0;
  else
	  d(5, 5) = pow((2 * (1 - omega(1))*(1 - omega(2))) / (2 - omega(1) - omega(2)), 2)*d(5,5);
}



void fixortodam::nlstresses (long ipp, long im, long ido)
{
  long i;
  long ncompstr=Mm->ip[ipp].ncompstr; // number of stress/strain components in vector form

  vector omega(3), omegao(3), xeq(3);
  vector epsg(ncompstr), eps(ncompstr), sig(ncompstr), sig_g(ncompstr);
  matrix epst(3,3), sigt(3,3), d(6,6);
  matrix tmat(3,3);

  // actual strains in the global coord. system
  for (i=0; i<ncompstr; i++)
    epsg(i) = Mm->ip[ipp].strain[i];

  // assemble transformation matrix(3,3) for the given integration point
  Mm->loc_transf_mat(tmat, ipp);

  //  transformation of strains given in the global coordinate system to the material local coord. system.
  vector_tensor (epsg,epst,Mm->ip[ipp].ssst,strain);
  glmatrixtransf(epst, tmat);  
  tensor_vector(eps,epst,Mm->ip[ipp].ssst,strain);

  // eps contains vector representation of strains in local coord. system of the material
  // epst contains tensor representation of strains in local coord. system of the material

  // attained values of damage parameters from the last equilibrium state
  for (i=0; i<3; i++)
    omegao(i) = Mm->ip[ipp].eqother[ido+i];

  // compute equivalent displacements
  compute_eqdispl(ipp, epst, xeq);

  // compute actual damage parameters
  compute_dam(xeq, omegao, omega);

  // secant stiffnes matrix
  secstiffmat(d, ipp, omega);
  
  // compute actual stress vector in the local coord. system of the material
  // sig = D(omega) * eps
  mxv(d, eps, sig);
  
  // transforms stress vector in the local coord. system of the material to global coord. system
  vector_tensor (sig,sigt,Mm->ip[ipp].ssst,stress);
  lgmatrixtransf(sigt, tmat);  
  tensor_vector(sig_g,sigt,Mm->ip[ipp].ssst,stress);

  // store actual stress in the global coord. system to the integration point
  for (i=0; i<ncompstr; i++)
    Mm->ip[ipp].stress[i] = sig_g[i];

  
  // store actual damage parameters to the integration point
  for (i=0; i<3; i++)
    Mm->ip[ipp].other[ido+i] = omega(i);
  for (i=0; i<3; i++)
    Mm->ip[ipp].other[ido+3+i] = xeq(i);
}



void fixortodam::updateval (long ipp,long im,long ido)
{
  long i;

  for (i=0; i<6; i++)
    Mm->ip[ipp].eqother[ido+i] = Mm->ip[ipp].other[ido+i];
}
